from liltab.model.heterogenous_attributes_network import HeterogenousAttributesNetwork
from liltab.data.dataloaders import ComposedDataLoader, FewShotDataLoader
from liltab.data.datasets import PandasDataset
from liltab.train.trainer import HeterogenousAttributesNetworkTrainer

from copy import deepcopy


def test_lighting_wrapper(resources_path):
    dataloader_1 = ComposedDataLoader(
        [
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_1.csv"),
                4,
                6,
                n_episodes=10,
            ),
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_2.csv"),
                4,
                6,
                n_episodes=10,
            ),
        ],
        batch_size=10,
    )

    dataloader_2 = deepcopy(dataloader_1)
    dataloader_3 = deepcopy(dataloader_2)

    model = HeterogenousAttributesNetwork()

    trainer = HeterogenousAttributesNetworkTrainer(
        n_epochs=1,
        gradient_clipping=True,
        learning_rate=1e-3,
        weight_decay=0.1,
        file_logger=False,
        tb_logger=False,
        model_checkpoints=False,
    )
    trainer.train_and_test(model, dataloader_1, dataloader_2, dataloader_3)



from __future__ import annotations

from torch import nn, Tensor
from typing import Callable


class NetworkBlock(nn.Module):
    """
    Module representing single neural network block.
    It's composed with Linear + Activation function
    and dropout.
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        activation_function: Callable,
        dropput_rate: float,
    ):
        """
        Args:
            input_size (int): Dimensionality of input.
            output_size (int): Dimensionality of output.
            activation_function (Callable): Activation function
                from torch.nn.
            dropput_rate (float): Dropout rate inside block.
        """
        super().__init__()

        self.block = nn.Sequential(
            nn.Linear(input_size, output_size),
            nn.Dropout(dropput_rate),
            activation_function,
        )

    def forward(self, X: Tensor) -> Tensor:
        return self.block(X)


class FeedForwardNetwork(nn.Module):
    """
    Module representing simple feed-forward neural network
    with only fully-connected layers.
    """

    def __init__(
        self,
        input_size: int,
        output_size: int,
        n_hidden_layers: int,
        hidden_size: int,
        dropout_rate: float,
        inner_activation_function: Callable = nn.ReLU(),
        output_activation_function: Callable = nn.Identity(),
    ):
        """
        Args:
            input_size (int): Dimensionality of input.
            output_size (int): Dimensionality of output.
            n_hidden_layers (int): Number of hidden layers.
            hidden_size (int): Size of hidden layer.
                Each layer has equal size.
            dropout_rate (float): Dropout rate for each hidden layer.
                Each layer has equal dropout rate.
            inner_activation_function (Callable, optional): Activation function
                from torch.nn which will be used after each hidden layer.
                Defaults to nn.ReLU().
            output_activation_function (Callable, optional):
            Activation function from torch.nn which will
            be used after output layer. Defaults to nn.Identity().
        """
        super().__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.n_hidden_layers = n_hidden_layers
        self.hidden_size = hidden_size
        self.dropout_rate = dropout_rate
        self.inner_activation_function = inner_activation_function
        self.output_activation_function = output_activation_function

        self.input_layer = NetworkBlock(
            self.input_size,
            self.hidden_size,
            self.inner_activation_function,
            self.dropout_rate,
        )
        self.hidden_layers = nn.ModuleList(
            [
                NetworkBlock(
                    self.hidden_size,
                    self.hidden_size,
                    self.inner_activation_function,
                    self.dropout_rate,
                )
                for _ in range(self.n_hidden_layers)
            ]
        )
        self.output_layer = NetworkBlock(
            self.hidden_size,
            self.output_size,
            self.output_activation_function,
            self.dropout_rate,
        )

    def forward(self, X: Tensor) -> Tensor:
        X = self.input_layer(X)
        for layer in self.hidden_layers:
            X = layer(X)
        X = self.output_layer(X)
        return X



import pytorch_lightning as pl

from pytorch_lightning.callbacks import Callback, ModelCheckpoint, EarlyStopping
from pathlib import Path
from torch import nn
from datetime import datetime
from typing import Optional, Callable, Union

from liltab.model.heterogenous_attributes_network import HeterogenousAttributesNetwork
from liltab.data.dataloaders import (
    ComposedDataLoader,
    RepeatableOutputComposedDataLoader,
)
from .utils import LightningWrapper
from .logger import TensorBoardLogger, FileLogger


class HeterogenousAttributesNetworkTrainer:
    """
    Class used for traning HeterogenousAttributesNetwork.
    """

    def __init__(
        self,
        n_epochs: int,
        gradient_clipping: bool,
        learning_rate: float,
        weight_decay: float,
        early_stopping: bool = False,
        loss: Callable = nn.MSELoss(),
        file_logger: bool = True,
        tb_logger: bool = True,
        model_checkpoints: bool = True,
        results_path: Union[str, Path, None] = None,
    ):
        """
        Args:
            n_epochs (int): number of epochs to train
            gradient_clipping (bool): If true, then gradient clipping is applied
            learning_rate (float): learning rate during training.
            weight_decay (float): weight decay during training.
            early_stopping (Optional, bool): if True, then early stopping with
                patience n_epochs // 10 is applied. Defaults to False.
            loss (Callable): Loss used during training. Defaults to MSELoss().
            file_logger (bool): if True, then file logger will write to
                {results_path} directory
            tb_logger (bool): if True, then tensorboard logger will write to
                {results_path} directory
            model_checkpoints (bool): if True, then model checkpoints will
                be loaded to {results_path} directory
            results_path (Optional, str, Path): directory to save logs and
                model checkpoints; required only if any of `file_logger`,
                `tb_logger`, `model_checkpoints` is not None
        """

        if not results_path and (file_logger or tb_logger or model_checkpoints):
            raise ValueError(
                """`results_path` is required if any of (`file_logger`,
                `tb_logger`, `model_checkpoints`) is not None"""
            )

        callbacks = []

        timestamp = datetime.now().strftime("%m-%d-%Y-%H:%M:%S")

        if file_logger:
            file_logger_callback = FileLogger(
                save_dir=Path(results_path) / timestamp, version="flat"
            )
        else:
            file_logger_callback = None

        if tb_logger:
            tb_logger_callback = TensorBoardLogger(
                save_dir=Path(results_path) / timestamp, version="tensorboard"
            )
        else:
            tb_logger_callback = None

        if file_logger or tb_logger:
            loggers_callback = LoggerCallback(
                file_logger=file_logger_callback, tb_logger=tb_logger_callback
            )
            callbacks.append(loggers_callback)

        if early_stopping:
            early_stopping = EarlyStopping(
                monitor="val_loss",
                mode="min",
                patience=100,
                min_delta=1e-3,
            )
            callbacks.append(early_stopping)

        if model_checkpoints:
            model_checkpoints_callback = ModelCheckpoint(
                dirpath=Path(results_path) / timestamp / "model_checkpoints",
                filename="model-{epoch}-{val_loss:.2f}",
                save_top_k=1,
                mode="min",
                every_n_epochs=10,
                save_last=True,
            )
            callbacks.append(model_checkpoints_callback)

        check_val_every_n_epoch = n_epochs // 1000 if n_epochs > 1000 else 1

        self.trainer = pl.Trainer(
            max_epochs=n_epochs,
            gradient_clip_val=1 if gradient_clipping else 0,
            check_val_every_n_epoch=check_val_every_n_epoch,
            callbacks=callbacks,
        )
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.loss = loss

    def train_and_test(
        self,
        model: HeterogenousAttributesNetwork,
        train_loader: ComposedDataLoader | RepeatableOutputComposedDataLoader,
        val_loader: ComposedDataLoader | RepeatableOutputComposedDataLoader,
        test_loader: ComposedDataLoader | RepeatableOutputComposedDataLoader,
    ) -> tuple[LightningWrapper, list[dict[str, float]]]:
        """
        Method used to train and test model.

        Args:
            model (HeterogenousAttributesNetwork): model to train
            train_loader (ComposedDataLoader | RepeatableOutputComposedDataLoader):
                loader withTrainingData
            val_loader (ComposedDataLoader | RepeatableOutputComposedDataLoader):
                loader with validation data
            test_loader (ComposedDataLoader | RepeatableOutputComposedDataLoader):
                loader with test data

        Returns:
            tuple[HeterogenousAttributesNetwork, list[dict[str, float]]]:
                trained network with metrics on test set.
        """
        model_wrapper = LightningWrapper(
            model,
            learning_rate=self.learning_rate,
            weight_decay=self.weight_decay,
            loss=self.loss,
        )
        self.trainer.fit(model_wrapper, train_loader, val_loader)
        test_results = self.trainer.test(model_wrapper, test_loader)
        return model_wrapper, test_results


class LoggerCallback(Callback):
    def __init__(
        self,
        file_logger: Optional[FileLogger] = None,
        tb_logger: Optional[TensorBoardLogger] = None,
    ) -> None:
        """
        Args:
            file_logger (FileLogger|None): csv logger
            tb_logger (TensorBoardLogger|None): tensorboard logger
        """
        super().__init__()
        self.file_logger = file_logger
        self.tb_logger = tb_logger

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        loss_value = outputs["loss"]
        pl_module.log("train_loss", loss_value, prog_bar=True, logger=False)

        if self.file_logger is not None:
            self.file_logger.log_train_value(loss_value)

        if self.tb_logger is not None:
            self.tb_logger.log_train_value(loss_value)

    def on_validation_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ) -> None:
        loss_value = outputs
        pl_module.log("val_loss", loss_value, prog_bar=True, logger=False)

        if self.file_logger is not None:
            self.file_logger.log_validate_value(loss_value)

        if self.tb_logger is not None:
            self.tb_logger.log_validate_value(loss_value)

    def on_test_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx=0
    ) -> None:
        loss_value = outputs
        pl_module.log("test_loss", loss_value, prog_bar=True, logger=False)

        if self.file_logger is not None:
            self.file_logger.log_test_value(loss_value)

        if self.tb_logger is not None:
            self.tb_logger.log_test_value(loss_value)



import torch

from liltab.train.logger import FileLogger, TensorBoardLogger


def test_file_logger_init_folders(results_path):
    logger_dir = results_path / "csv_logger"
    version = "test_1"
    _ = FileLogger(save_dir=logger_dir, version=version)

    test_file = logger_dir / version / "test.csv"
    train_file = logger_dir / version / "train.csv"
    validate_file = logger_dir / version / "validate.csv"

    assert test_file.exists()
    assert train_file.exists()
    assert validate_file.exists()


def test_tb_logger_init_folders(results_path):
    logger_dir = results_path / "tb_logger"
    version = "test"
    logger = TensorBoardLogger(save_dir=logger_dir, version=version)
    logger.log_train_value(torch.Tensor([1]))

    # dictionary created after first call
    logger_full_dir = logger_dir / version

    assert logger_full_dir.exists()


def test_file_logger_insert_to_file(results_path):
    logger_dir = results_path / "csv_logger"
    version = "test_2"
    logger = FileLogger(save_dir=logger_dir, version=version)
    train_file = logger_dir / version / "train.csv"

    values_inserted = 99

    for _ in range(values_inserted):
        logger.log_train_value(torch.Tensor([1]))

    with open(train_file, "r") as f:
        assert len(f.readlines()) == values_inserted



from .datasets import PandasDataset
from .dataloaders import FewShotDataLoader, ComposedDataLoader, RepeatableOutputComposedDataLoader
from pathlib import Path
from typing import Any, Type


class ComposedDataLoaderFactory:
    """
    Factory which simplifies creation of ComposedDataLoader.
    """

    @staticmethod
    def create_composed_dataloader_from_path(
        path: Path,
        dataset_cls: Type = PandasDataset,
        dataset_creation_args: dict[str, Any] = None,
        loader_cls: Type = FewShotDataLoader,
        dataloader_creation_args: dict[str, Any] = None,
        composed_dataloader_cls: Type = ComposedDataLoader,
        batch_size: int = 32,
    ) -> ComposedDataLoader | RepeatableOutputComposedDataLoader:
        """
        Creates composed data loader using directory with csv files. Each file is wrapped in
        dataset selected dataset_cls. Next these datasets are wrapped in dataloader
        specified in loader_cls and finally it is composed to dataset specified in
        composed_dataloader_cls.

        Args:
            path (Path): path to data stored in csv format. Should contain only csv.
            dataset_cls (Type, optional): Class which will encapsulate csv data to torch dataset.
                Defaults to PandasDataset.
            dataset_creation_args (dict[str, Any], optional): Arguments passed to dataset
            constructors. Defaults to None.
            loader_cls (Type, optional): Class which will be used to encapsulate datasets.
                Defaults to FewShotDataLoader.
            dataloader_creation_args (dict[str, Any], optional): Arguments passed to dalaoader
                constructors. Defaults to None.
            composed_dataloader_cls (Type, optional): Class encapsulating all created dataloaders.
                Defaults to ComposedDataLoader.
            batch_size (int, optional): size of batch which created dataloader will return.
                Defaults to 32.

        Returns:
            ComposedDataLoader | RepeatableOutputComposedDataLoader:
                Dataloader created using aforementioned algorithm and parameters.
        """
        if not dataset_creation_args:
            dataset_creation_args = {}
        if not dataloader_creation_args:
            dataloader_creation_args = {}

        pandas_datasets = [dataset_cls(file, **dataset_creation_args) for file in path.iterdir()]
        loaders = [loader_cls(dataset, **dataloader_creation_args) for dataset in pandas_datasets]
        return composed_dataloader_cls(loaders, batch_size)



import numpy as np

from copy import deepcopy
from torch import Tensor
from torch.utils.data import Dataset
from typing import Iterable, OrderedDict


class FewShotDataLoader:
    """
    DataLoader, which iterates dataset in few-shot learning manner
    i.e. when called next(loader) it returns episode with
    properly sampled support and query sets.
    """

    def __init__(
        self,
        dataset: Dataset,
        support_size: int,
        query_size: int,
        n_episodes: int = None,
    ):
        """
        Args:
            dataset (Dataset): dataset to load data from.
            support_size (int): size of support set in each episode.
            query_size (int): size of query set in each episode.
            n_episodes (int, optional): number of episodes.
                If none, then iterator is without end. Defaults to None.
        """
        self.dataset = dataset
        self.support_size = support_size
        self.query_size = query_size
        self.n_episodes = n_episodes

        self.curr_episode = 0

        self.n_rows = len(self.dataset)

    def __iter__(self):
        return deepcopy(self)

    def __next__(self) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        """
        Returns episode i.e. support and query sets
        with sized specified in constructors.

        Returns:
            tuple[Tensor, Tensor, Tensor, Tensor]:
                (X_support, y_support, X_query, y_query)
        """
        if self.n_episodes:
            if self.curr_episode == self.n_episodes:
                raise StopIteration()
            self.curr_episode += 1

        replace = True if self.support_size + self.query_size >= self.n_rows else False
        all_drawn_indices = np.random.choice(
            self.n_rows, self.support_size + self.query_size, replace=replace
        )
        support_indices = np.random.choice(all_drawn_indices, self.support_size, replace=False)
        query_indices = np.array(list(set(all_drawn_indices) - set(support_indices)))
        return *self.dataset[support_indices], *self.dataset[query_indices]

    def has_next(self) -> bool:
        return self.curr_episode != self.n_episodes


class ComposedDataLoader:
    """
    DataLoader which wraps list of data loaders objects and
    when next(dataloader) is called, then returns episode from
    randomly chosen one of passed dataloaders.
    """

    def __init__(
        self,
        dataloaders: list[Iterable],
        batch_size: int = 32,
        num_batches: int = 1,
    ):
        """
        Args:
            dataloaders (list[Iterable]): list of
                dataloaders to sample from
            batch_size (int, optional): size of batch.
                Defaults to 32.
            batch_size (int, optional): number of returned batches.
                Defaults to 1.
        """
        self.dataloaders = dataloaders
        self.batch_size = batch_size

        self.counter = 0
        self.num_batches = num_batches
        self.n_dataloaders = len(dataloaders)

    def __iter__(self):
        return deepcopy(self)

    def __next__(self):
        if self.counter == self.num_batches:
            raise StopIteration()
        self.counter += 1
        return [self._get_single_example() for _ in range(self.batch_size)]

    def _get_single_example(self) -> tuple[Tensor, Tensor, Tensor, Tensor]:
        """
        Returns support and query sets from one DataLoaders from
        randomly chosen from passed dataloaders.

        Returns:
            tuple[Tensor, Tensor, Tensor, Tensor]:
                (X_support, y_support, X_query, y_query)
        """
        dataloader_has_next = False
        while not dataloader_has_next:
            dataloader_idx = np.random.choice(self.n_dataloaders, 1)[0]
            dataloader = self.dataloaders[dataloader_idx]
            dataloader_has_next = dataloader.has_next()
        return next(dataloader)


class RepeatableOutputComposedDataLoader:
    """
    DataLoader which wraps list of data loaders objects
    and takes one observation from each of them and when
    next(dataloader) is called it returns always the same batch
    of data. Useful with test/validation datasets.
    """

    def __init__(self, dataloaders: list[Iterable], *args, **kwargs):
        """
        Args:
            dataloaders (list[Iterable]): list of
                dataloaders to sample from.
        """
        self.dataloaders = dataloaders

        self.batch_counter = 0
        self.n_dataloaders = len(dataloaders)

        self.loaded = False
        self.cache = OrderedDict()
        for i, dataloader in enumerate(self.dataloaders):
            self.cache[i] = next(dataloader)

    def __iter__(self):
        return deepcopy(self)

    def __next__(self):
        if self.loaded:
            raise StopIteration()
        self.loaded = True
        return [sample for _, sample in self.cache.items()]



import numpy as np

from liltab.data.datasets import PandasDataset
from liltab.data.dataloaders import (
    FewShotDataLoader,
    ComposedDataLoader,
    RepeatableOutputComposedDataLoader,
)


def test_few_shot_data_loader_returns_proper_episodes_count(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    n_episodes = 100
    dataset = PandasDataset(frame_path)
    dataloader = FewShotDataLoader(dataset, 2, 8, n_episodes)

    episodes = list(dataloader)
    assert len(episodes) == n_episodes


def test_few_shot_data_loader_returns_proper_shape_tensors(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    dataset = PandasDataset(frame_path)
    dataloader = FewShotDataLoader(dataset, 4, 6)
    n_features = len(dataset.feature_columns)

    X_support, y_support, X_query, y_query = next(dataloader)

    assert X_support.shape == (4, n_features)
    assert y_support.shape == (4, 1)
    assert X_query.shape == (6, n_features)
    assert y_query.shape == (6, 1)


def test_few_shot_data_loader_returns_disjoint_tensors(resources_path, utils):
    frame_path = resources_path / "random_df_1.csv"
    dataset = PandasDataset(frame_path)
    dataloader = FewShotDataLoader(dataset, 4, 6, n_episodes=10)

    for episode in dataloader:
        X_support, y_support, X_query, y_query = episode

        assert not utils.tensors_have_common_rows(X_support, X_query)
        assert not utils.tensors_have_common_rows(y_support, y_query)


def test_few_shot_data_loader_has_next(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    dataset = PandasDataset(frame_path)
    dataloader = FewShotDataLoader(dataset, 4, 6, n_episodes=10)

    assert dataloader.has_next()
    for _ in range(10):
        next(dataloader)
    assert not dataloader.has_next()


def test_composed_data_loader_returns_batch_with_proper_size(resources_path):
    dataloader = ComposedDataLoader(
        [
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_1.csv"),
                4,
                6,
                n_episodes=10,
            ),
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_2.csv"),
                4,
                6,
                n_episodes=10,
            ),
        ],
        batch_size=10,
    )
    loaded_dataset = list(dataloader)[0]
    assert len(loaded_dataset) == 10


def test_composed_data_loader_returns_proper_batch_count(resources_path):
    dataloader = ComposedDataLoader(
        [
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_1.csv"),
                4,
                6,
                n_episodes=100,
            ),
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_2.csv"),
                4,
                6,
                n_episodes=100,
            ),
        ],
        batch_size=10,
        num_batches=5,
    )
    loaded_dataset = list(dataloader)
    assert len(loaded_dataset) == 5


def test_composed_data_loader_returns_from_all_loaders_properly(
    resources_path,
):
    n_episodes = 11
    dataloader = ComposedDataLoader(
        [
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_1.csv"),
                4,
                6,
                n_episodes=4,
            ),
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_2.csv"),
                4,
                6,
                n_episodes=7,
            ),
        ],
        batch_size=n_episodes,
    )
    loaded_dataset = list(dataloader)
    batches_lens = np.array(list(map(lambda t: t[0][1].shape[0], loaded_dataset[0])))

    assert (batches_lens == 4).sum() == 4
    assert (batches_lens == 2).sum() == 7


def test_composed_data_loader_returns_from_all_loaders_almost_equally(
    resources_path,
):
    def run_experiment():
        batch_size = 32
        dataloader = ComposedDataLoader(
            [
                FewShotDataLoader(
                    PandasDataset(resources_path / "random_df_1.csv"),
                    4,
                    6,
                    n_episodes=100,
                ),
                FewShotDataLoader(
                    PandasDataset(resources_path / "random_df_2.csv"),
                    4,
                    6,
                    n_episodes=100,
                ),
            ],
            batch_size=batch_size,
        )
        batch = list(dataloader)[0]
        batches_lens = np.array(list(map(lambda t: t[0][1].shape[0], batch)))
        return [(batches_lens == 4).sum(), (batches_lens == 2).sum()]

    experiments_results = np.array([run_experiment() for _ in range(20)]).mean(axis=0)
    assert (experiments_results >= 14).all()


def test_repeatable_output_composed_data_loader_repeat_samples(resources_path, utils):
    dataloader = RepeatableOutputComposedDataLoader(
        [
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_1.csv"),
                4,
                6,
                n_episodes=100,
            ),
            FewShotDataLoader(
                PandasDataset(resources_path / "random_df_2.csv"),
                4,
                6,
                n_episodes=100,
            ),
        ]
    )

    dataloader_1 = list(dataloader)[0]
    dataloader_2 = list(dataloader)[0]

    for sample_1, sample_2 in zip(dataloader_1, dataloader_2):
        X_support_1, y_support_1, X_query_1, y_query_1 = sample_1
        X_support_2, y_support_2, X_query_2, y_query_2 = sample_2
        assert utils.tensors_have_common_rows(X_support_1, X_support_2)
        assert utils.tensors_have_common_rows(y_support_1, y_support_2)
        assert utils.tensors_have_common_rows(X_query_1, X_query_2)
        assert utils.tensors_have_common_rows(y_query_1, y_query_2)



import yaml
import pytorch_lightning as pl

from pytorch_lightning.callbacks import ModelCheckpoint
from datetime import datetime

from liltab.data.datasets import PandasDataset
from liltab.data.dataloaders import (
    FewShotDataLoader,
    ComposedDataLoader,
    RepeatableOutputComposedDataLoader,
)
from liltab.data.factory import ComposedDataLoaderFactory
from liltab.model.heterogenous_attributes_network import HeterogenousAttributesNetwork
from liltab.train.trainer import HeterogenousAttributesNetworkTrainer
from liltab.train.logger import TensorBoardLogger, FileLogger
from loguru import logger
from pathlib import Path
from torch import nn


def main():
    config_path = Path("config/03_openml_clf_data_experiment_config.yaml")
    logger_type = "both"
    use_profiler = "no"

    pl.seed_everything(123)

    logger.info("Loading config")
    with open(config_path) as f:
        config = yaml.load(f, Loader=yaml.CLoader)

    logger.info("Loading data")
    train_loader = ComposedDataLoaderFactory.create_composed_dataloader_from_path(
        Path(config["train_data_path"]),
        PandasDataset,
        {"encode_categorical_target": True},
        FewShotDataLoader,
        {"support_size": config["support_size"], "query_size": config["query_size"]},
        ComposedDataLoader,
        batch_size=config["batch_size"],
    )
    val_loader = ComposedDataLoaderFactory.create_composed_dataloader_from_path(
        Path(config["val_data_path"]),
        PandasDataset,
        {"encode_categorical_target": True},
        FewShotDataLoader,
        {"support_size": config["support_size"], "query_size": config["query_size"]},
        RepeatableOutputComposedDataLoader,
        batch_size=config["batch_size"],
    )
    test_loader = ComposedDataLoaderFactory.create_composed_dataloader_from_path(
        Path(config["test_data_path"]),
        PandasDataset,
        {"encode_categorical_target": True},
        FewShotDataLoader,
        {"support_size": config["support_size"], "query_size": config["query_size"]},
        RepeatableOutputComposedDataLoader,
        batch_size=config["batch_size"],
    )

    logger.info("Creating model")
    model = HeterogenousAttributesNetwork(
        hidden_representation_size=config["hidden_representation_size"],
        n_hidden_layers=config["n_hidden_layers"],
        hidden_size=config["hidden_size"],
        dropout_rate=config["dropout_rate"],
        is_classifier=config["is_classifier"],
    )

    results_path = Path("results") / config["name"]

    trainer = HeterogenousAttributesNetworkTrainer(
        n_epochs=config["num_epochs"],
        gradient_clipping=config["gradient_clipping"],
        learning_rate=config["learning_rate"],
        weight_decay=config["weight_decay"],
        early_stopping=config["early_stopping"],
        loss=nn.CrossEntropyLoss(),
        file_logger=True,
        tb_logger=True,
        model_checkpoints=True,
        results_path=results_path,
    )

    logger.info("Training model")
    trainer.train_and_test(
        model=model,
        train_loader=train_loader,
        val_loader=val_loader,
        test_loader=test_loader,
    )


if __name__ == "__main__":
    main()



import torch
import torch.nn.functional as F

from torch import nn, Tensor
from typing import Callable
from .utils import FeedForwardNetwork


class HeterogenousAttributesNetwork(nn.Module):
    """
    Module representing neural network wchich takes as an input:
        * support set attributes
        * support set responses
        * query set attributes
    And returns responses corresponding to query set.
    Support and query set must have same attributes.
    Can take any number of attributes.
    """

    def __init__(
        self,
        hidden_representation_size: int = 32,
        n_hidden_layers: int = 3,
        hidden_size: int = 32,
        dropout_rate: int = 0.1,
        inner_activation_function: Callable = nn.ReLU(),
        output_activation_function: Callable = nn.Identity(),
        is_classifier: bool = False,
    ):
        """
        Args:
            hidden_representation_size (int, optional): Size of hidden
                representation sizes i. e. all intermediate network outputs.
                Defaults to 32.
            n_hidden_layers (int, optional): Number hidden layers of networks
                used during inference. Defaults to 3.
            hidden_size (int, optional): Number of neurons per hidden layer
                in networks using during inference. Defaults to 32.
            dropout_rate (int, optional): Dropout rate of networks
                used during inference. Defaults to 0.1.
            inner_activation_function (Callable, optional): Inner activation function
                of networks used during inference. Should be function from torch.nn.
                Defaults to ReLU.
            output_activation_function (Callable, optional): Output activation function
                of networks used during inference. Should be function from torch.nn.
                Defaults to nn.Identity().
            is_classifier (bool, optional): If true then output of the network will
                generate probabilities of classes for query set Defaults to False.
        """
        super().__init__()
        self.initial_features_encoding_network = FeedForwardNetwork(
            1,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )
        self.initial_features_representation_network = FeedForwardNetwork(
            hidden_representation_size,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )
        self.interaction_encoding_network = FeedForwardNetwork(
            hidden_representation_size + 1,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )
        self.interaction_representation_network = FeedForwardNetwork(
            hidden_representation_size,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )

        self.features_encoding_network = FeedForwardNetwork(
            hidden_representation_size + 1,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )
        self.features_representation_network = FeedForwardNetwork(
            hidden_representation_size,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )

        self.inference_encoding_network = FeedForwardNetwork(
            hidden_representation_size + 1,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )
        self.inference_embedding_network = FeedForwardNetwork(
            hidden_representation_size,
            hidden_representation_size,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            inner_activation_function,
        )
        self.inference_network = FeedForwardNetwork(
            2 * hidden_representation_size,
            1,
            n_hidden_layers,
            hidden_size,
            dropout_rate,
            inner_activation_function,
            output_activation_function,
        )
        self.is_classifier = is_classifier

    def forward(self, X_support: Tensor, y_support: Tensor, X_query: Tensor) -> Tensor:
        """
        Inference function of network. Inference is done in following steps:
            1. Calculate initial representation for all atrributes and responses
                in support set.
            2. Calculate representations for all observations in support set
                i. e. calculate representation of support set using aforementioned
                initial representations of attributes and responses
            3. Calculate final attributes and responses representations in support set
            4. Calculate representation of query set.attributes.
            5. Make prediction based on representations of support set attributes and
                responses and query set attributes representations.
        All representations calculations are done using feed forward neural networks.
        Args:
            X_support (Tensor): Support set attributes
                with shape (n_support_observations, n_attributes)
            y_support (Tensor): Support set responses
                with shape (n_support_observations, n_responses)
            X_query (Tensor): Query set attributes
                with shape (n_query_observations, n_attributes)
        Returns:
            Tensor: Inferred query set responses shaped (n_query_obervations, n_responses)
        """
        attributes_initial_representation = self._calculate_initial_features_representation(
            self.initial_features_encoding_network,
            self.initial_features_representation_network,
            X_support,
        )
        responses_initial_representation = self._calculate_initial_features_representation(
            self.initial_features_encoding_network,
            self.initial_features_representation_network,
            y_support,
        )
        support_set_representation = self._calculate_support_set_representation(
            self.interaction_encoding_network,
            self.interaction_representation_network,
            X_support,
            attributes_initial_representation,
            y_support,
            responses_initial_representation,
        )
        attributes_representation = self._calculate_features_representation(
            self.features_encoding_network,
            self.features_representation_network,
            support_set_representation,
            X_support,
        )
        responses_representation = self._calculate_features_representation(
            self.features_encoding_network,
            self.features_representation_network,
            support_set_representation,
            y_support,
        )

        if self.is_classifier:
            prediction = self._make_prediction_clf(
                self.inference_encoding_network,
                self.inference_embedding_network,
                attributes_representation,
                X_support,
                X_query,
                y_support,
            )
        else:
            prediction = self._make_prediction_reg(
                self.inference_encoding_network,
                self.inference_embedding_network,
                self.inference_network,
                attributes_representation,
                responses_representation,
                X_query,
            )

        return prediction

    def _calculate_initial_features_representation(
        self,
        encoder_network: FeedForwardNetwork,
        representation_network: FeedForwardNetwork,
        set_: Tensor,
    ) -> Tensor:
        """
        Calculates initial features (can be attributes or resposnes) of given set
        using feed forward neural network.
        Args:
            encoder_network (FeedForwardNetwork): Feed forward network used to encode
                features. Network takes as an input single number and returns
                vector with length hidden_representation_size.
            representation_network (FeedForwardNetwork): Feed forward network used to
                calculate initial representation of features. Network takes as an
                input vector with length hidden_representation_size + 1 and returns
                vector with length hidden_representation_size.
            X (Tensor): set itself. Can be only attributes or only responses.
                Tensor with shape (n_observations, n_features).
        Returns:
            Tensor: calculated initial features representation with shape
                (n_features, hidden_representation_size) containing representations
                of attributes in rows.
        """
        initial_tensor_shape = set_.shape
        representation_length = encoder_network.output_size
        network_input = set_.reshape(-1, 1)
        encoded_input = encoder_network(network_input)
        encoded_input = encoded_input.reshape(*initial_tensor_shape, representation_length)
        encoded_input = encoded_input.mean(axis=0)
        features_representation = representation_network(encoded_input)
        return features_representation

    def _calculate_support_set_representation(
        self,
        interaction_encoding_network: FeedForwardNetwork,
        interaction_representation_network: FeedForwardNetwork,
        X: Tensor,
        attributes_initial_representation: Tensor,
        y: Tensor,
        responses_initial_representation: Tensor,
    ) -> Tensor:
        """
        Calculates representation of observations in given support set
        using feed-forwand neural networks.
        Args:
            interaction_encoding_network (FeedForwardNetwork): Feed forward neural network
                used to encode interactions between features representations and observations
                in X. Takes as an input vector with length hidden_representation_size + 1
                and returns vector with length hidden_representation_size.
            interaction_representation_network (FeedForwardNetwork): Feed forward neural network
                used to calculate representation of interactions between features
                representations and observations in X.
                Takes as an input vector with length hidden_representation_size + 1
                and returns vector with length hidden_representation_size.
            X (Tensor): attributes of support set with shape (n_support_observations, n_attributes).
            attributes_initial_representation (Tensor): initial representation of attributes
                of support set with shape (n_attributes, hidden_representation_size).
            y (Tensor): responses of support set with shape (n_support_observations, n_responses).
            responses_initial_representation (Tensor): initial representation of responses
                of support set with shape (n_responses, hidden_representation_size).
        Returns:
            Tensor: calcualted representaion of support selfset
                with shape (n_support_observations, hidden_representation_size).
        """
        attributes_encoded = self._calculate_interaction_encoding(
            interaction_encoding_network, attributes_initial_representation, X
        )
        responses_encoded = self._calculate_interaction_encoding(
            interaction_encoding_network, responses_initial_representation, y
        )

        return interaction_representation_network(attributes_encoded + responses_encoded)

    def _calculate_interaction_encoding(
        self,
        interaction_encoding_network: FeedForwardNetwork,
        representation: Tensor,
        set_: Tensor,
    ) -> Tensor:
        """
        Method used to calculate encoding of interaction encoding
        between initial representation of features in set and
        observations in this set.
        Args:
            interaction_encoding_network (FeedForwardNetwork): Feed forward neural network
                used to encode interactions between features representations and observations
                in X. Takes as an input vector with length hidden_representation_size + 1
                and returns vector with length hidden_representation_size.
            representation (Tensor): initial representation of features
                of support set with shape (n_features, hidden_representation_size).
            set_ (Tensor): set containing observations to calculate interations encoding with.
                Can be only attributes or only features.
                Tensor with shape (n_observations, n_features).
        Returns:
            Tensor: Calculated interations encoding
                with shape (n_observations, hidden_representation_size).
        """
        encoding_input = self._enrich_representation_with_set_rows(representation, set_)
        representation_with_set = encoding_input.reshape(-1, representation.shape[1] + 1)
        encoding = interaction_encoding_network(representation_with_set)
        encoding = encoding.reshape(*set_.shape, interaction_encoding_network.output_size).mean(
            axis=1
        )
        return encoding

    def _calculate_features_representation(
        self,
        features_encoding_network: FeedForwardNetwork,
        features_representation_network: FeedForwardNetwork,
        set_representation: Tensor,
        set_: Tensor,
    ) -> Tensor:
        """
        Method used to calculate representations of features based on
        set representation and observations from set.
        Args:
            features_encoding_network (FeedForwardNetwork): Feed forward neural
                network used for calculation of encoding of concatenation
                of initial representation and features from set_.
                Takes as an input vector with length hidden_representation_size + 1
                and returns vector with length hidden_representation_size.
            features_representation_network (FeedForwardNetwork): Feed forward neural
                network used for calculation of representation of concatenation
                of calculated encoding.
                Takes as an input vector with length hidden_representation_size
                and returns vector with length hidden_representation_size.
            set_representation (Tensor): representation of set with shape
                (n_observations, hidden_representation_size).
            set_ (Tensor): set itself. Can be only attributes or only features.
                Tensor with shape (n_observations, n_features).
        Returns:
            Tensor: Calculated representation of features
                with shape (n_features, hidden_representation_size).
        """
        n_features = set_.shape[1]
        attributes_encoding_input = self._enrich_representation_with_set_features(
            set_representation, set_
        )
        features_encoded = features_encoding_network(
            attributes_encoding_input.reshape(-1, set_representation.shape[1] + 1)
        )
        features_encoded = features_encoded.reshape(
            n_features, -1, set_representation.shape[1]
        ).mean(axis=1)
        features_representation = features_representation_network(features_encoded)
        return features_representation

    def _enrich_representation_with_set_features(
        self, representation: Tensor, set_: Tensor
    ) -> Tensor:
        """
        Calculates tensor which is repeated representation with feature
        from set as a last column for each repetition e.g.
        representation:
            [
                [0.1, 0.2, 0.3],
                [0.4, 0.5, 0.6],
                [0.7, 0.8, 0.9]
            ]
        set_:
            [
                [1, 2],
                [3, 4],
                [5, 6]
            ]
        result:
            [
                [ # First feature
                    [0.1, 0.2, 0.3, 1],
                    [0.4, 0.5, 0.6, 3],
                    [0.7, 0.8, 0.9, 5]
                ],
                [ # Second feature
                    [0.1, 0.2, 0.3, 2],
                    [0.4, 0.5, 0.6, 4],
                    [0.7, 0.8, 0.9, 6]
                ]
            ]
        Args:
            representation (Tensor): representation of feautres
                with shape (n_features, hidden_representation_size)
            set_ (Tensor): set to enrich representation with
                with shape (n_observations, n_features)
        Returns:
            Tensor: enriched representations with shape
                (n_observations, n_features, hidden_representation_size + 1)
        """
        n_examples, n_features = set_.shape
        repeated_representation = representation.repeat(n_features, 1).reshape(
            n_features, n_examples, -1
        )
        representation_with_set = torch.concat(
            [
                repeated_representation,
                torch.unsqueeze(set_.T, 2),
            ],
            axis=2,
        )
        return representation_with_set

    def _make_prediction_reg(
        self,
        inference_encoding_network: FeedForwardNetwork,
        inference_embedding_network: FeedForwardNetwork,
        inference_network: FeedForwardNetwork,
        attributes_representation: Tensor,
        responses_representation: Tensor,
        X_query: Tensor,
    ) -> Tensor:
        """
        Method responsible for calculation of final prediction based on support
        set representation and query set in regression problem.
        Args:
            inference_encoding_network (FeedForwardNetwork): Network responsible
                for encoding concatenation of support set attributes
                representation with query set attributes.
                Takes as an input vector with length hidden_representation_size + 1
                and returns vector with length hidden_representation_size.
            inference_embedding_network (FeedForwardNetwork): Network responsible
                for calculating representation of encoding created by
                inference_encoding_network.
                Takes as an input vector with length hidden_representation_size
                and returns vector with length hidden_representation_size.
            inference_network (FeedForwardNetwork): Network responsible
                for calculation of responses corresponding to query set based
                on representation of attributes and responses from query set
                and representation of query set observations.
                Takes as an input vector with length 2*hidden_representation_size
                and returns vector with length n_responses.
            attributes_representation (Tensor): representation of attributes of support set
                with shape (n_attributes, hidden_representation_size)
            responses_representation (Tensor): representation of responses of support set
                with shape (n_responses, hidden_representation_size)
            X_query (Tensor): query set with shape (n_query_observations, n_attributes)
        Returns:
            Tensor: responses corresponding to X_query
                with shape (n_query_observations, n_responses)
        """
        X_query_inference_embedding = self._get_inference_embedding_of_set(
            inference_encoding_network,
            inference_embedding_network,
            X_query,
            attributes_representation,
        )
        response_dim = responses_representation.shape[0]
        n_query_observations = X_query.shape[0]
        query_embedding_dim = X_query_inference_embedding.shape[1]
        attributes_representation_dim = attributes_representation.shape[1]

        inference_network_input = (
            X_query_inference_embedding.repeat(response_dim, 1)
            .reshape(response_dim, n_query_observations, inference_encoding_network.output_size)
            .transpose(1, 0)
        )

        responses_representation_expanded = (
            responses_representation.unsqueeze(1)
            .repeat(n_query_observations, 1, 1)
            .reshape(n_query_observations, response_dim, -1)
        )
        inference_network_input = torch.concat(
            [
                responses_representation_expanded,
                inference_network_input,
            ],
            axis=2,
        )
        inference_network_input = inference_network_input.reshape(
            -1, query_embedding_dim + attributes_representation_dim
        )
        response = inference_network(inference_network_input)
        response = response.reshape(n_query_observations, -1)
        return response

    def _make_prediction_clf(
        self,
        inference_encoding_network: FeedForwardNetwork,
        inference_embedding_network: FeedForwardNetwork,
        attributes_representation: Tensor,
        X_support: Tensor,
        X_query: Tensor,
        y_support: Tensor,
    ) -> Tensor:
        """
        Method responsible for calculation of final prediction based on support
        set representation and query set in classification problem.
        Args:
            inference_encoding_network (FeedForwardNetwork): Network responsible
                for encoding concatenation of support set attributes
                representation with query set attributes.
                Takes as an input vector with length hidden_representation_size + 1
                and returns vector with length hidden_representation_size.
            inference_embedding_network (FeedForwardNetwork): Network responsible
                for calculating representation of encoding created by
                inference_encoding_network.
                Takes as an input vector with length hidden_representation_size
                and returns vector with length hidden_representation_size.
            inference_network (FeedForwardNetwork): Network responsible
                for calculation of responses corresponding to query set based
                on representation of attributes and responses from query set
                and representation of query set observations.
                Takes as an input vector with length 2*hidden_representation_size
                and returns vector with length n_responses.
            attributes_representation (Tensor): representation of attributes of support set
                with shape (n_attributes, hidden_representation_size)
            responses_representation (Tensor): representation of responses of support set
                with shape (n_responses, hidden_representation_size)
            X_query (Tensor): query set with shape (n_query_observations, n_attributes)
            X_support (Tensor): support set with shape (n_support_observations, n_attributes)
        Returns:
            Tensor: responses corresponding to X_query
                with shape (n_query_observations, n_responses)
        """
        X_support_inference_embedding = self._get_inference_embedding_of_set(
            inference_encoding_network,
            inference_embedding_network,
            X_support,
            attributes_representation,
        )
        X_query_inference_embedding = self._get_inference_embedding_of_set(
            inference_encoding_network,
            inference_embedding_network,
            X_query,
            attributes_representation,
        )

        classes_representations = self._calculate_classes_representations(
            X_support_inference_embedding, y_support
        )
        response = F.softmax(
            -(torch.cdist(X_query_inference_embedding, classes_representations) ** 2), dim=1
        )

        return response

    def _calculate_classes_representations(self, X: Tensor, y: Tensor) -> Tensor:
        """
        Calculates classes representations by averaging their observations.

        Args:
            X (Tensor): Observations.
            y (Tensor): Categorical responses, one-hot encoded.

        Returns:
            Tensor: Calculated representations ordered by corresponding response value.
        """
        response_values = torch.arange(y.shape[1])
        y = y.argmax(axis=1)
        classes_representations = torch.zeros((response_values.shape[0], X.shape[1]))
        for val in response_values:
            if (y == val).sum() != 0:
                classes_representations[val] = X[y == val].mean(axis=0)
        return classes_representations

    def _get_inference_embedding_of_set(
        self,
        inference_encoding_network: FeedForwardNetwork,
        inference_embedding_network: FeedForwardNetwork,
        set_: Tensor,
        attributes_representation: Tensor,
    ) -> Tensor:
        """
        Calculates embedding of set used during prediction generation
        based on attributes representation.

        Args:
            inference_encoding_network (FeedForwardNetwork): Network responsible
                for encoding concatenation of support set attributes
                representation with query set attributes.
                Takes as an input vector with length hidden_representation_size + 1
                and returns vector with length hidden_representation_size.
            inference_embedding_network (FeedForwardNetwork): Network responsible
                for calculating representation of encoding created by
                inference_encoding_network.
                Takes as an input vector with length hidden_representation_size
                and returns vector with length hidden_representation_size.
            set_ (Tensor): Set to generate embedding with shape
                (n_observations, n_attributes)
            attributes_representation (Tensor): representation of attributes of support set
                with shape (n_attributes, hidden_representation_size)

        Returns:
            Tensor: calculated embedding with shape (n_observations, hidden_representation_size)
        """
        X_query_encoding_input = self._enrich_representation_with_set_rows(
            attributes_representation, set_
        )
        X_query_encoding_input = X_query_encoding_input.reshape(
            -1, attributes_representation.shape[1] + 1
        )

        X_query_inference_embedding = inference_encoding_network(X_query_encoding_input)
        X_query_inference_embedding = X_query_inference_embedding.reshape(*set_.shape, -1).mean(
            axis=1
        )
        X_query_inference_embedding = inference_embedding_network(X_query_inference_embedding)
        return X_query_inference_embedding

    def _enrich_representation_with_set_rows(self, representation: Tensor, set_: Tensor) -> Tensor:
        """
        Calculates tensor which is repeated representation with observaion
        from set as a last column for each repetition e.g.
        representation:
            [
                [0.1, 0.2, 0.3],
                [0.4, 0.5, 0.6]
            ]
        set_:
            [
                [1, 2],
                [3, 4],
                [5, 6]
            ]
        result:
            [
                [ # First observation
                    [0.1, 0.2, 0.3, 1],
                    [0.4, 0.5, 0.6, 2]
                ],
                [ # Second observation
                    [0.1, 0.2, 0.3, 3],
                    [0.4, 0.5, 0.6, 4]
                ],
                [ # Third observation
                    [0.1, 0.2, 0.3, 5],
                    [0.4, 0.5, 0.6, 6]
                ]
            ]
        Args:
            representation (Tensor): representation of feautres
                with shape (n_features, hidden_representation_size)
            set_ (Tensor): set to enrich representation with
                with shape (n_observations, n_features)
        Returns:
            Tensor: enriched representations with shape
                (n_observations, n_features, hidden_representation_size + 1)
        """
        n_rows, n_features = set_.shape
        repeated_representation = representation.repeat([n_rows, 1]).reshape(n_rows, n_features, -1)
        representation_with_set = torch.concat(
            [repeated_representation, torch.unsqueeze(set_, 2)], 2
        )
        return representation_with_set



import typer
import yaml
import pytorch_lightning as pl
import warnings

from liltab.data.datasets import PandasDataset, RandomFeaturesPandasDataset
from liltab.data.dataloaders import (
    FewShotDataLoader,
    ComposedDataLoader,
    RepeatableOutputComposedDataLoader,
)
from liltab.data.factory import ComposedDataLoaderFactory
from liltab.model.heterogenous_attributes_network import HeterogenousAttributesNetwork
from liltab.train.trainer import HeterogenousAttributesNetworkTrainer
from liltab.train.logger import TensorBoardLogger, FileLogger
from loguru import logger
from typing_extensions import Annotated
from pathlib import Path

warnings.filterwarnings("ignore")
app = typer.Typer()


@app.command(help="Trains network on heterogenous attribute spaces.")
def main(
    config_path: Annotated[Path, typer.Option(..., help="Path to experiment configuration.")],
    logger_type: Annotated[
        str,
        typer.Option(
            ...,
            help="""typer of logger. tb=[tensorboard],
            flat=[flat file], both=[tensoboard and flat file]""",
        ),
    ] = "both",
    use_profiler: Annotated[
        str,
        typer.Option(
            ...,
            help="""""use profiler (take long time, 8-10 epoches suggested),
            yes or no; requires tensorboard (logger-type=[tb|both])""",
        ),
    ] = "no",
    seed: Annotated[int, typer.Option(..., help="Seed")] = 123,
):
    pl.seed_everything(seed)

    logger.info("Loading config")
    with open(config_path) as f:
        config = yaml.load(f, Loader=yaml.CLoader)

    logger.info("Loading data")
    train_loader = ComposedDataLoaderFactory.create_composed_dataloader_from_path(
        Path(config["train_data_path"]),
        RandomFeaturesPandasDataset,
        {},
        FewShotDataLoader,
        {"support_size": config["support_size"], "query_size": config["query_size"]},
        ComposedDataLoader,
        batch_size=config["batch_size"],
    )
    val_loader = ComposedDataLoaderFactory.create_composed_dataloader_from_path(
        Path(config["val_data_path"]),
        PandasDataset,
        {},
        FewShotDataLoader,
        {"support_size": config["support_size"], "query_size": config["query_size"]},
        RepeatableOutputComposedDataLoader,
        batch_size=config["batch_size"],
    )
    test_loader = ComposedDataLoaderFactory.create_composed_dataloader_from_path(
        Path(config["test_data_path"]),
        PandasDataset,
        {},
        FewShotDataLoader,
        {"support_size": config["support_size"], "query_size": config["query_size"]},
        RepeatableOutputComposedDataLoader,
        batch_size=config["batch_size"],
    )

    logger.info("Creating model")
    model = HeterogenousAttributesNetwork(
        hidden_representation_size=config["hidden_representation_size"],
        n_hidden_layers=config["n_hidden_layers"],
        hidden_size=config["hidden_size"],
        dropout_rate=config["dropout_rate"],
    )

    if logger_type == "tb":
        tb_logger = TensorBoardLogger(
            "results/tensorboard",
            name=config["name"],
            use_profiler=True if use_profiler == "yes" else False,
        )
        file_logger = None
    elif logger_type == "flat":
        tb_logger = None
        file_logger = FileLogger("results/flat")
    elif logger_type == "both":
        tb_logger = TensorBoardLogger(
            "results/tensorboard",
            name=config["name"],
            use_profiler=True if use_profiler == "yes" else False,
        )
        file_logger = FileLogger("results/flat")
    else:
        raise ValueError("logger_type must from [tb, flat, both]")

    trainer = HeterogenousAttributesNetworkTrainer(
        n_epochs=config["num_epochs"],
        gradient_clipping=config["gradient_clipping"],
        learning_rate=config["learning_rate"],
        weight_decay=config["weight_decay"],
        early_stopping=config["early_stopping"],
        file_logger=file_logger,
        tb_logger=tb_logger,
    )

    logger.info("Training model")
    trainer.train_and_test(
        model=model,
        train_loader=train_loader,
        val_loader=val_loader,
        test_loader=test_loader,
    )


if __name__ == "__main__":
    app()



from sklearn.pipeline import make_pipeline
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler


def get_preprocessing_pipeline() -> Pipeline:
    """
    Returns preprocessing pipeline composed with:
        1. Simple imputer with mean strategy
        2. StandardScaler
    Returns:
        Pipeline: Pipeline with preprocessing steps.
    """
    return make_pipeline(SimpleImputer(strategy="mean"), StandardScaler())



import numpy as np
import torch
import torch.nn.functional as F

from liltab.model.heterogenous_attributes_network import HeterogenousAttributesNetwork
from torch import Tensor, nn, optim, testing
from torchtest import assert_vars_change


def test_network_initializes_properly():
    network = HeterogenousAttributesNetwork(
        hidden_representation_size=20,
        n_hidden_layers=5,
        hidden_size=20,
        dropout_rate=0.1,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    assert network.initial_features_encoding_network.input_size == 1
    assert network.initial_features_encoding_network.output_size == 20
    assert network.initial_features_encoding_network.n_hidden_layers == 5
    assert network.initial_features_encoding_network.hidden_size == 20
    assert network.initial_features_encoding_network.dropout_rate == 0.1

    assert network.initial_features_representation_network.input_size == 20
    assert network.initial_features_representation_network.output_size == 20
    assert network.initial_features_representation_network.n_hidden_layers == 5
    assert network.initial_features_representation_network.hidden_size == 20
    assert network.initial_features_representation_network.dropout_rate == 0.1

    assert network.interaction_encoding_network.input_size == 21
    assert network.interaction_encoding_network.output_size == 20
    assert network.interaction_encoding_network.n_hidden_layers == 5
    assert network.interaction_encoding_network.hidden_size == 20
    assert network.interaction_encoding_network.dropout_rate == 0.1

    assert network.interaction_representation_network.input_size == 20
    assert network.interaction_representation_network.output_size == 20
    assert network.interaction_representation_network.n_hidden_layers == 5
    assert network.interaction_representation_network.hidden_size == 20
    assert network.interaction_representation_network.dropout_rate == 0.1

    assert network.features_encoding_network.input_size == 21
    assert network.features_encoding_network.output_size == 20
    assert network.features_encoding_network.n_hidden_layers == 5
    assert network.features_encoding_network.hidden_size == 20
    assert network.features_encoding_network.dropout_rate == 0.1

    assert network.features_representation_network.input_size == 20
    assert network.features_representation_network.output_size == 20
    assert network.features_representation_network.n_hidden_layers == 5
    assert network.features_representation_network.hidden_size == 20
    assert network.features_representation_network.dropout_rate == 0.1

    assert network.inference_encoding_network.input_size == 21
    assert network.inference_encoding_network.output_size == 20
    assert network.inference_encoding_network.n_hidden_layers == 5
    assert network.inference_encoding_network.hidden_size == 20
    assert network.inference_encoding_network.dropout_rate == 0.1

    assert network.inference_network.input_size == 40
    assert network.inference_network.output_size == 1
    assert network.inference_network.n_hidden_layers == 5
    assert network.inference_network.hidden_size == 20
    assert network.inference_network.dropout_rate == 0.1


def test_forward_returns_proper_shape(utils):
    X_support = Tensor(np.random.uniform(size=(5, 10)))
    y_support = Tensor(np.random.uniform(size=(5, 3)))
    X_query = Tensor(np.random.uniform(size=(27, 10)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0.1,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    prediction = network(X_support, y_support, X_query)

    assert prediction.shape == (27, 3)


def test_forward_returns_probabilities_when_classifier():
    X_support = Tensor(np.random.uniform(size=(5, 10)))
    y_support = Tensor([[1, 0, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 0]])
    X_query = Tensor(np.random.uniform(size=(27, 10)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0.1,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
        is_classifier=True,
    )

    prediction = network(X_support, y_support, X_query)

    assert prediction.shape == (X_query.shape[0], y_support.shape[1])
    assert (prediction >= 0).all()
    testing.assert_close(prediction.sum(axis=1), torch.ones(27))


def test_all_network_params_are_trained(utils):
    X_support = Tensor(np.random.uniform(size=(5, 10)))
    y_support = Tensor(np.random.uniform(size=(5, 3)))
    X_query = Tensor(np.random.uniform(size=(27, 10)))
    y_query = Tensor(np.random.uniform(size=(27, 3)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0.1,
        inner_activation_function=nn.ELU(),
        output_activation_function=nn.Identity(),
    )

    inference_adapter = utils.get_inference_adapter(network, X_support, y_support)

    assert_vars_change(
        model=inference_adapter,
        loss_fn=F.cross_entropy,
        optim=optim.Adam(inference_adapter.parameters()),
        batch=(X_query, y_query),
        device="cpu:0",
    )


def test_calculate_initial_features_representation():
    X_support = Tensor(np.random.uniform(size=(5, 10)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    actual_initial_features_representation = network._calculate_initial_features_representation(
        network.initial_features_encoding_network,
        network.initial_features_representation_network,
        X_support,
    )

    expected_initial_features_representation_example = X_support[:, 1].unsqueeze(1)
    expected_initial_features_representation_example = network.initial_features_encoding_network(
        expected_initial_features_representation_example
    ).mean(axis=0)
    expected_initial_features_representation_example = (
        network.initial_features_representation_network(
            expected_initial_features_representation_example
        )
    )

    assert actual_initial_features_representation.shape == (10, 32)
    testing.assert_close(
        actual_initial_features_representation[1], expected_initial_features_representation_example
    )


def test_calculate_support_set_representation():
    X_support = Tensor(np.random.uniform(size=(5, 10)))
    attributes_initial_representation = Tensor(np.random.uniform(size=(10, 32)))
    y_support = Tensor(np.random.uniform(size=(5, 3)))
    responses_initial_representation = Tensor(np.random.uniform(size=(3, 32)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    actual_representation_shape = network._calculate_support_set_representation(
        network.features_encoding_network,
        network.features_representation_network,
        X_support,
        attributes_initial_representation,
        y_support,
        responses_initial_representation,
    ).shape

    assert actual_representation_shape == (5, 32)


def test_calculate_interaction_encoding():
    X_support = Tensor(np.random.uniform(size=(5, 10)))
    features_representation = Tensor(np.random.uniform(size=(10, 32)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    actual = network._calculate_interaction_encoding(
        network.interaction_encoding_network, features_representation, X_support
    )
    expected_example = network._enrich_representation_with_set_rows(
        features_representation, X_support
    )[1]
    expected_example = network.interaction_encoding_network(expected_example)
    expected_example = expected_example.mean(axis=0)

    testing.assert_close(actual[1], expected_example)


def test_calculate_features_representation():
    X_support = Tensor(np.random.uniform(size=(5, 10)))
    set_representation = Tensor(np.random.uniform(size=(5, 32)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    actual_representation_shape = network._calculate_features_representation(
        network.features_encoding_network,
        network.features_representation_network,
        set_representation,
        X_support,
    ).shape

    assert actual_representation_shape == (10, 32)


def test_enrich_representation_with_set_features():
    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    representation = Tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]])
    set_ = Tensor([[1, 2], [3, 4], [5, 6]])

    expected = Tensor(
        [
            [[0.1, 0.2, 0.3, 1], [0.4, 0.5, 0.6, 3], [0.7, 0.8, 0.9, 5]],
            [[0.1, 0.2, 0.3, 2], [0.4, 0.5, 0.6, 4], [0.7, 0.8, 0.9, 6]],
        ]
    )
    actual = network._enrich_representation_with_set_features(representation, set_)

    testing.assert_close(actual, expected)


def test_make_prediction_reg():
    attributes_representation = Tensor(np.random.uniform(size=(10, 32)))
    responses_representation = Tensor(np.random.uniform(size=(3, 32)))
    X_query = Tensor(np.random.uniform(size=(5, 10)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    actual_prediction = network._make_prediction_reg(
        network.inference_encoding_network,
        network.inference_embedding_network,
        network.inference_network,
        attributes_representation,
        responses_representation,
        X_query,
    )

    query_example_embedding = X_query[1].unsqueeze(0)
    query_example_embedding = network._enrich_representation_with_set_rows(
        attributes_representation, query_example_embedding
    )
    query_example_embedding = query_example_embedding.reshape(-1, 33)
    query_example_embedding = network.inference_encoding_network(query_example_embedding).mean(
        axis=0
    )
    query_example_embedding = network.inference_embedding_network(query_example_embedding)

    inference_network_example_input = query_example_embedding.repeat(3, 1)
    inference_network_example_input = torch.concat(
        [responses_representation, inference_network_example_input], axis=1
    )
    expected_example_response = network.inference_network(inference_network_example_input).reshape(
        3
    )

    assert actual_prediction.shape == (5, 3)
    testing.assert_close(actual_prediction[1], expected_example_response)


def test_get_inference_embedding_of_set():
    attributes_representation = Tensor(np.random.uniform(size=(10, 32)))
    X_query = Tensor(np.random.uniform(size=(5, 10)))

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    expected_query_example_embedding = X_query[1].unsqueeze(0)
    expected_query_example_embedding = network._enrich_representation_with_set_rows(
        attributes_representation, expected_query_example_embedding
    )
    expected_query_example_embedding = expected_query_example_embedding.reshape(10, 33)
    expected_query_example_embedding = network.inference_encoding_network(
        expected_query_example_embedding
    ).mean(axis=0)
    expected_query_example_embedding = network.inference_embedding_network(
        expected_query_example_embedding
    )

    actual_query_example_embedding = network._get_inference_embedding_of_set(
        network.inference_encoding_network,
        network.inference_embedding_network,
        X_query,
        attributes_representation,
    )

    testing.assert_close(actual_query_example_embedding[1], expected_query_example_embedding)


def test_calculate_classes_representations():
    X = torch.Tensor([[1, 2, 3], [4, 5, 6], [6, 5, 4], [3, 2, 1]])
    y = torch.Tensor([[0, 1], [1, 0], [1, 0], [0, 1]])
    expected_representations = Tensor([[5.0, 5.0, 5.0], [2.0, 2.0, 2.0]])

    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    actual_representations = network._calculate_classes_representations(X, y)

    testing.assert_close(expected_representations, actual_representations)


def test_enrich_representation_with_set_rows():
    network = HeterogenousAttributesNetwork(
        hidden_representation_size=32,
        n_hidden_layers=3,
        hidden_size=32,
        dropout_rate=0,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Identity(),
    )

    representation = Tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
    set_ = Tensor([[1, 2], [3, 4], [5, 6]])

    expected = Tensor(
        [
            [[0.1, 0.2, 0.3, 1], [0.4, 0.5, 0.6, 2]],
            [[0.1, 0.2, 0.3, 3], [0.4, 0.5, 0.6, 4]],
            [[0.1, 0.2, 0.3, 5], [0.4, 0.5, 0.6, 6]],
        ]
    )
    actual = network._enrich_representation_with_set_rows(representation, set_)

    testing.assert_close(actual, expected)



import numpy as np
import torch.nn.functional as F

from torch import nn, optim, Tensor
from torchtest import assert_vars_change
from liltab.model.utils import NetworkBlock, FeedForwardNetwork


def test_network_block_initializes_properly():
    block = NetworkBlock(
        input_size=12, output_size=30, activation_function=nn.ReLU(), dropput_rate=0.45
    )

    assert len(block.block) == 3
    assert block.block[1].p == 0.45


def test_network_block_accepts_and_returns_propoer_shape():
    block = NetworkBlock(
        input_size=12, output_size=30, activation_function=nn.ReLU(), dropput_rate=0.45
    )
    input = Tensor(np.random.uniform(size=(3, 12)))
    output = block(input)

    assert output.shape == (3, 30)


def test_network_block_is_trained():
    block = NetworkBlock(
        input_size=12, output_size=30, activation_function=nn.ReLU(), dropput_rate=0.45
    )
    batch = (Tensor(np.random.uniform(size=(3, 12))), Tensor(np.random.uniform(size=(3, 30))))

    assert_vars_change(
        model=block,
        loss_fn=F.cross_entropy,
        optim=optim.Adam(block.parameters()),
        batch=batch,
        device="cpu:0",
    )


def test_feed_forward_network_initializes_properly():
    network = FeedForwardNetwork(
        input_size=12,
        output_size=30,
        n_hidden_layers=5,
        hidden_size=20,
        dropout_rate=0.5,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Sigmoid(),
    )

    assert len(network.hidden_layers) == 5

    assert network.input_layer.block[0].in_features == 12
    assert network.input_layer.block[0].out_features == 20
    for block in network.hidden_layers:
        assert block.block[0].in_features == 20
        assert block.block[0].out_features == 20
    assert network.output_layer.block[0].in_features == 20
    assert network.output_layer.block[0].out_features == 30


def test_feed_forward_network_accepts_and_returns_propoer_shape():
    network = FeedForwardNetwork(
        input_size=12,
        output_size=30,
        n_hidden_layers=5,
        hidden_size=20,
        dropout_rate=0.5,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Sigmoid(),
    )
    input = Tensor(np.random.uniform(size=(10, 12)))
    output = network(input)
    assert output.shape == (10, 30)


def test_feed_forward_network_is_trained():
    network = FeedForwardNetwork(
        input_size=12,
        output_size=30,
        n_hidden_layers=5,
        hidden_size=20,
        dropout_rate=0.5,
        inner_activation_function=nn.ReLU(),
        output_activation_function=nn.Sigmoid(),
    )
    batch = (Tensor(np.random.uniform(size=(10, 12))), Tensor(np.random.uniform(size=(10, 30))))

    assert_vars_change(
        model=network,
        loss_fn=F.cross_entropy,
        optim=optim.Adam(network.parameters()),
        batch=batch,
        device="cpu:0",
    )



import numpy as np
import pandas as pd
import pytest

from liltab.data.datasets import PandasDataset, RandomFeaturesPandasDataset
from numpy.testing import assert_almost_equal
from torch import Tensor, float32


def test_dataset_initializes_default_columns(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    df = pd.read_csv(frame_path)
    frame_columns = df.columns.tolist()

    dataset = PandasDataset(frame_path)

    assert dataset.feature_columns == frame_columns[:-1]
    assert dataset.target_columns == [frame_columns[-1]]


def test_dataset_assigns_non_default_columns(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    df = pd.read_csv(frame_path)
    frame_columns = df.columns.tolist()

    dataset = PandasDataset(
        frame_path,
        feature_columns=frame_columns[1:3],
        target_columns=frame_columns[4:],
    )

    assert dataset.feature_columns == frame_columns[1:3]
    assert dataset.target_columns == frame_columns[4:]


def test_indexing_dataset_returns_proper_data(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    df = pd.read_csv(frame_path)
    index = [2, 3, 6, 8, 10, 11]

    dataset = PandasDataset(frame_path, preprocess_data=False)

    expected_records = df.loc[index]
    actual_X, actual_y = dataset[index]

    assert_almost_equal(actual_X.numpy(), expected_records[dataset.feature_columns].values)
    assert_almost_equal(actual_y.numpy(), expected_records[dataset.target_columns].values)


def test_indexing_dataset_returns_proper_data_with_preprocessing(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    df = pd.read_csv(frame_path)
    df = (df - df.mean(axis=0)) / df.std(axis=0)
    index = [2, 3, 6, 8, 10, 11]

    dataset = PandasDataset(frame_path)

    expected_records = df.loc[index]
    actual_X, actual_y = dataset[index]

    assert_almost_equal(
        actual_X.numpy(), expected_records[dataset.feature_columns].values, decimal=2
    )
    assert_almost_equal(
        actual_y.numpy(), expected_records[dataset.target_columns].values, decimal=2
    )


def test_class_forbids_one_hot_with_multiple_targets(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    df = pd.read_csv(frame_path)
    feture_columns = df.columns[:-2]
    target_columns = df.columns[-2:]
    with pytest.raises(ValueError):
        PandasDataset(frame_path, feture_columns, target_columns, encode_categorical_target=True)


def test_preprocessing_when_target_categorical(resources_path):
    frame_path = resources_path / "random_df_3.csv"
    df = pd.read_csv(frame_path)
    expected_X = df.drop(columns=["class"])
    expected_X = (expected_X - expected_X.mean(axis=0)) / expected_X.std(axis=0)

    dataset = PandasDataset(frame_path, encode_categorical_target=True)

    assert dataset.y.shape == (df.shape[0], df["class"].max())
    assert_almost_equal(dataset.y.sum(axis=1).numpy(), np.ones(df.shape[0]))
    assert_almost_equal(dataset.X.numpy(), expected_X.values, decimal=2)


def test_indexing_returns_dataset_with_proper_type(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    index = [1, 2, 3]
    dataset = PandasDataset(frame_path)

    X, y = dataset[index]

    assert type(X) is Tensor
    assert X.dtype == float32
    assert type(y) is Tensor
    assert y.dtype == float32


def test_random_features_pandas_dataset_has_no_null(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    dataset = RandomFeaturesPandasDataset(frame_path, preprocess_data=True)
    assert dataset.df.isnull().sum().sum() == 0


def test_random_features_pandas_dataset_is_standarized(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    dataset = RandomFeaturesPandasDataset(frame_path, preprocess_data=True)
    assert (np.abs(dataset.df.std() - 1) < 1e-1).all()
    assert (np.abs(dataset.df.mean()) < 1e-6).all()


def test_random_features_pandas_dataset_change_features(resources_path):
    frame_path = resources_path / "random_df_1.csv"
    persist_features_iter = 3
    dataset = RandomFeaturesPandasDataset(
        frame_path, preprocess_data=True, persist_features_iter=persist_features_iter
    )

    previous_features = np.ndarray([])
    previous_target = np.ndarray([])

    features_change_cnt = 0
    target_change_cnt = 0

    for _ in range(int(1e2)):
        dataset[0]
        features_change_cnt += int(not np.array_equal(previous_features, dataset.features))
        target_change_cnt += int(not np.array_equal(previous_target, dataset.target))

        previous_features = dataset.features
        previous_target = dataset.target

    assert np.abs(features_change_cnt / int(1e2) - 1 / persist_features_iter) < 2e-1
    assert np.abs(target_change_cnt / int(1e2) - 1 / persist_features_iter) < 2e-1



import numpy as np
import pandas as pd
import torch

from pathlib import Path
from torch import Tensor
from torch.utils.data import Dataset

from .preprocessing import get_preprocessing_pipeline


class PandasDataset(Dataset):
    """
    Torch wrapper to pandas DataFrame which makes it usable
    as torch.Dataset. Data from this class is returned
    as Tensor.
    """

    def __init__(
        self,
        data_path: Path,
        feature_columns: list[str] = None,
        target_columns: list[str] = None,
        preprocess_data: bool = True,
        encode_categorical_target: bool = False,
    ):
        """
        Args:
            data_path (Path): Path to data to be loaded
            feature_columns (list[str], optional): Columns from frame
                which will be used as features.
                Defaults to all columns without last.
            target_columns (list[str], optional): Columns from frame
                to be used as features. Defaults to last column from frame.
            preprocess_data (bool, optional): If true, then imputes data
                using mean strategy and standardizes using StandardScaler.
                Defaults to True.
            encode_categorical_target(bool, optional): if True, then target column
                will be encoded using one-hot. Works only with single target variable.
                Default to False.
        """
        self.data_path = data_path
        self.encode_categorical_target = encode_categorical_target
        self.df = pd.read_csv(data_path)
        self.feature_columns = (
            feature_columns if feature_columns is not None else self.df.columns.tolist()[:-1]
        )
        self.target_columns = (
            target_columns if target_columns is not None else [self.df.columns.tolist()[-1]]
        )

        if len(self.target_columns) > 1 and self.encode_categorical_target:
            raise ValueError("One-hot encoding is supported only for single target")

        if preprocess_data:
            self.preprocessing_pipeline = get_preprocessing_pipeline()
            if self.encode_categorical_target:
                self.df.loc[:, self.feature_columns] = self.preprocessing_pipeline.fit_transform(
                    self.df[self.feature_columns]
                )
            else:
                self.df = pd.DataFrame(
                    self.preprocessing_pipeline.fit_transform(self.df), columns=self.df.columns
                )
        self.X = torch.from_numpy(self.df[self.feature_columns].to_numpy()).type(torch.float32)

        self.y = self.df[self.target_columns]
        if self.encode_categorical_target:
            self.y = pd.get_dummies(self.y.astype("category"))
        self.y = torch.from_numpy(self.y.to_numpy()).type(torch.float32)

    def __getitem__(self, idx: list[int]) -> tuple[Tensor, Tensor]:
        X = self.X[idx]
        y = self.y[idx]

        return X, y

    def __len__(self) -> int:
        return self.df.shape[0]


class RandomFeaturesPandasDataset(Dataset):
    """
    Torch wrapper to pandas DataFrame which makes it usable
    as torch.Dataset. Every persist_features_iter iteration
    object selects random attributes and responses columns
    which will be returned as Tensor.
    """

    def __init__(
        self,
        data_path: Path,
        persist_features_iter: int = 2,
        preprocess_data: bool = True,
    ):
        """
        Args:
            data_path (Path): Path to data to be loaded
            persist_features_iter (int, optional): For how many
                iterations persist current selection of features.
                Defaults to 2.
            preprocess_data(bool, optional): If true, then imputes data
                using mean strategy and standardizes using StandardScaler.
                Defaults to True.
        """
        self.data_path = data_path
        self.persist_features_iter = persist_features_iter

        self.df = pd.read_csv(data_path)
        self.columns = self.df.columns.values
        self.n_columns = len(self.columns)

        if preprocess_data:
            self.preprocessing_pipeline = get_preprocessing_pipeline()
            self.df = pd.DataFrame(
                self.preprocessing_pipeline.fit_transform(self.df), columns=self.df.columns
            )

        self.persist_features_counter = 0
        self.features = None
        self.target = None

    def __getitem__(self, idx: list[int]) -> tuple[Tensor, Tensor]:
        if self.persist_features_counter == 0:
            self.persist_features_counter = self.persist_features_iter
            col_idx = np.arange(self.n_columns)
            features_size = np.random.randint(low=1, high=self.n_columns)
            features_idx = np.random.choice(col_idx, features_size)
            remaining_idx = list(set(col_idx) - set(features_idx))
            target_idx = np.random.choice(remaining_idx, 1)
            self.features, self.target = self.columns[features_idx], self.columns[target_idx]
        self.persist_features_counter -= 1

        X = torch.from_numpy(self.df[self.features].to_numpy()).type(torch.float32)
        y = torch.from_numpy(self.df[self.target].to_numpy()).type(torch.float32)

        return X[idx], y[idx]

    def __len__(self) -> int:
        return self.df.shape[0]



import pytorch_lightning as pl
import torch
import torch.nn.functional as F

from torch import optim, Tensor
from typing import Any, Callable

from ..model.heterogenous_attributes_network import HeterogenousAttributesNetwork


class LightningWrapper(pl.LightningModule):
    """
    Wrapper around pyTorch model, which makes it compatible with pyTorch-lightning
    framework. It's purpose is to implement pyTorch-lightning hooks.
    """

    def __init__(
        self,
        model: HeterogenousAttributesNetwork,
        learning_rate: float,
        weight_decay: float,
        loss: Callable = F.mse_loss,
    ):
        """
        Args:
            model (HeterogenousAttributesNetwork): network to be wrapped around.
            learning_rate (float): learning rate during training.
            weight_decay (float): weight decay during training.
            loss (Callable): loss function used during training.
        """
        super().__init__()
        self.model = model
        self.loss = loss
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.metrics_history = dict()

        self.save_hyperparameters()

    def training_step(self, batch: list[tuple[Tensor, Tensor, Tensor, Tensor]], batch_idx) -> float:
        sum_loss_value = 0.0
        for i, example in enumerate(batch):
            X_support, y_support, X_query, y_query = example
            prediction = self.model(X_support, y_support, X_query)
            loss = self.loss(prediction, y_query)
            if torch.isnan(loss):
                sum_loss_value = sum_loss_value * (i + 1) / i if i > 0 else 0
            else:
                sum_loss_value += loss

        return sum_loss_value

    def validation_step(
        self, batch: list[tuple[Tensor, Tensor, Tensor, Tensor]], batch_idx
    ) -> float:
        sum_loss_value = 0.0
        for i, example in enumerate(batch):
            X_support, y_support, X_query, y_query = example
            prediction = self.model(X_support, y_support, X_query)
            loss = self.loss(prediction, y_query)
            if torch.isnan(loss):
                sum_loss_value = sum_loss_value * (i + 1) / i if i > 0 else 0
            else:
                sum_loss_value += loss

        return sum_loss_value

    def test_step(self, batch: list[tuple[Tensor, Tensor, Tensor, Tensor]], batch_idx) -> float:
        sum_loss_value = 0.0
        for i, example in enumerate(batch):
            X_support, y_support, X_query, y_query = example
            prediction = self.model(X_support, y_support, X_query)
            loss = self.loss(prediction, y_query)
            if torch.isnan(loss):
                sum_loss_value = sum_loss_value * (i + 1) / i if i > 0 else 0
            else:
                sum_loss_value += loss

        return sum_loss_value

    def configure_optimizers(self) -> Any:
        return optim.Adam(self.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay)

    def calculate_model_weights_norm(self) -> Tensor:
        pass



from liltab.model.heterogenous_attributes_network import HeterogenousAttributesNetwork
from torch import nn, Tensor
from pathlib import Path
from pytest import fixture
from shutil import rmtree


@fixture(scope="session")
def resources_path():
    return Path("test/resources")


@fixture(scope="session")
def results_path():
    return Path("test/results")


class Utils:
    @staticmethod
    def tensors_have_common_rows(t1: Tensor, t2: Tensor) -> bool:
        for row_t1 in t1:
            for row_t2 in t2:
                if ((row_t1 - row_t2).abs() < 1e-9).all():
                    return True
        return False

    @staticmethod
    def get_inference_adapter(network, X_support, y_support):
        return InferenceAdapter(network, X_support, y_support)


@fixture(scope="session")
def utils():
    return Utils


class InferenceAdapter(nn.Module):
    def __init__(
        self,
        network: HeterogenousAttributesNetwork,
        X_support: Tensor,
        y_support: Tensor,
    ):
        super().__init__()
        self.network = network
        self.X_support = X_support
        self.y_support = y_support

    def forward(self, X_query: Tensor):
        return self.network(self.X_support, self.y_support, X_query)


def pytest_sessionfinish(session, exitstatus):
    rmtree(Path("test") / "results")


def pytest_sessionstart(session):
    results_directory = Path("test") / "results"

    if results_directory.exists():
        rmtree(results_directory)

    results_directory.mkdir(parents=False, exist_ok=False)



from pathlib import Path
from datetime import datetime

from typing import Any
from lightning_fabric.utilities.types import _PATH
from pytorch_lightning.loggers import TensorBoardLogger as TBLogger


class TensorBoardLogger(TBLogger):
    def __init__(
        self,
        save_dir: _PATH = "results/tensorboard",
        version: str = None,
        name: str = "",
        **kwargs: Any,
    ):
        if version is None:
            _version = "experiment_" + name + " " + datetime.now().strftime("%m-%d-%Y-%H:%M:%S")
        else:
            _version = version

        self.iter = 0

        super().__init__(
            save_dir,
            name=None,
            version=_version,
            log_graph=False,
            default_hp_metric=False,
            prefix="",
            sub_dir=None,
            **kwargs,
        )

    def log_train_value(self, value: float) -> None:
        self.iter += 1
        self.experiment.add_scalar("train loss", value, self.iter)

    def log_test_value(self, value: float) -> None:
        self.experiment.add_text("test loss", str(value))

    def log_validate_value(self, value: float) -> None:
        self.experiment.add_scalar("validate loss", value, self.iter)

    def log_weights(self, weights) -> None:
        self.experiment.add_histogram("weights and biases", weights)


class FileLogger:
    def __init__(self, save_dir: str = "results/flat", name: str = "", version: str = None) -> None:
        super().__init__()
        self.save_dir = save_dir

        if version is None:
            self.version = "experiment " + name + " " + datetime.now().strftime("%m-%d-%Y-%H:%M:%S")
        else:
            self.version = version

        experiment_path = Path(save_dir)
        experiment_path = experiment_path / self.version
        experiment_path.mkdir(exist_ok=False, parents=True)

        self.train_loss_path = experiment_path / "train.csv"
        self.train_loss_path.touch(exist_ok=False)

        self.validate_loss_path = experiment_path / "validate.csv"
        self.validate_loss_path.touch(exist_ok=False)

        self.test_loss_path = experiment_path / "test.csv"
        self.test_loss_path.touch(exist_ok=False)

    def log_train_value(self, value: float) -> None:
        with open(self.train_loss_path, "a") as f:
            f.write(str(value.item()))
            f.write("\n")

    def log_test_value(self, value: float) -> None:
        with open(self.test_loss_path, "a") as f:
            f.write(str(value.item()))
            f.write("\n")

    def log_validate_value(self, value: float) -> None:
        with open(self.validate_loss_path, "a") as f:
            f.write(str(value.item()))
            f.write("\n")
